from judy_core.agents.llm_agent import LLMAgent
from judy_core.agents.types import ModelInputMessage


class StatefulLLMAgent(LLMAgent):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._messages: list[ModelInputMessage] = []
        self._tools = []
