from pydantic import BaseModel


class JudyCompletionOption(BaseModel):
    stream_chunk_timeout: float | None = 160
    num_retries: int = 5


class LitellmCompletionOption(BaseModel):
    # litellm默认超时时间是10分钟，太长了
    timeout: float = 150
    num_retries: int = 5
