from judy_core.agents.configuration import Configuration
from judy_core.agents.notification_center import NotificationCenter
from tongcell.core import ActorId

CONFIGURATION_ACTOR_ID = ActorId.default(Configuration.actor_kind)
NOTIFICATION_CENTER_ACTOR_ID = ActorId.default(NotificationCenter.actor_kind)

VERIFIER_NAME = "verifier-0"


def get_agent_name(prefix: str, suffix: str) -> str:
    return "{}-{}".format(prefix, suffix)


def get_executor_name(suffix: str):
    return get_agent_name("executor", suffix)
